/*
	*
	* This file is a part of CoreKeyboard.
	* An on-screenkeyboard for CuboCore Application Suite
	* Copyright 2019 Britanicus <marcusbritanicus@gmail.com>
	*

	*
	* This program is free software; you can redistribute it and/or modify
	* it under the terms of the GNU General Public License as published by
	* the Free Software Foundation; either version 3 of the License, or
	* (at your option) any later version.
	*

	*
	* This program is distributed in the hope that it will be useful,
	* but WITHOUT ANY WARRANTY; without even the implied warranty of
	* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	* GNU General Public License for more details.
	*

	*
	* You should have received a copy of the GNU General Public License
	* along with this program; if not, see {http://www.gnu.org/licenses/}.
	*
*/

#include "corekeyboard.h"
#include "settings.h"

#include <cprime/settingsmanage.h>

#include <QStringList>
#include <QVBoxLayout>
#include <QHBoxLayout>
#include <QDesktopWidget>
#include <QApplication>
#include <QSettings>
#include <QSystemTrayIcon>
#include <QScreen>
#include <QLabel>
#include <QX11Info>
#include <X11/extensions/XTest.h>
#include <X11/Xlocale.h>
#include <X11/Xos.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xproto.h>
#include <X11/XKBlib.h>

// X11 Headers
#include <X11/X.h>
#include "fixx11h.h"

settingsManage *sm = settingsManage::initialize( true);
QSettings *CoreKeyboard::keymap = nullptr;

const QString alphabet = "abcdefghijklmnopqrstuvwxyz-";

inline QWidget* blank() {
	QWidget *w = new QWidget();
	w->setMinimumSize( 1, 1 );
	w->setSizePolicy( QSizePolicy::MinimumExpanding, QSizePolicy::MinimumExpanding );

	return w;
}

CoreKeyboard::CoreKeyboard() : QMainWindow() {

	keymap = new QSettings( CKSettings::keyMapping(), QSettings::NativeFormat );
	dict = new Suggestions();

	maxRows = keymap->value( "Rows" ).toStringList().count();

	/* Extra row for suggestions */
	if ( CKSettings::showSuggestions() )
		maxRows += 1;

	/* Dirty test hack */
	if ( keymap->value( "Pages" ).toStringList().count() == 1 )
		maxCols = 15;

	if ( CKSettings::keyboardLayout() == CKSettings::Desktop and CKSettings::showExtraDesktopKeys() )
		maxCols += 3;

    createKeyboard();
    setWidgetProperties();

    trayIcon = new trayicon( this );
    connect( trayIcon, SIGNAL( toggleShowHide() ), this, SLOT( toggleShowHide() ) );

    setFont( QFont( CKSettings::buttonFontName(), 7, QFont::Bold ) );
    setMouseTracking( true );

   	if ( CKSettings::showSuggestions() )
		updatePredictions();
};

void CoreKeyboard::createKeyboard() {

	fm = new QFontMetrics( font() );

    QGridLayout *keyboardLyt = new QGridLayout();
    keyboardLyt->setContentsMargins( QMargins( 5, 5, 5, 5 ) );
    keyboardLyt->setSpacing( 1 );

	suggestions = new QScrollArea();
	suggestions->setFixedHeight( MINIMUM_BUTTON_SIZE );
	suggestions->setWidgetResizable( true );
	suggestions->setVerticalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
	suggestions->setHorizontalScrollBarPolicy( Qt::ScrollBarAlwaysOff );
	suggestions->setFrameStyle( QFrame::NoFrame );
	suggestions->setStyleSheet( "background-color: transparent;" );

	if ( CKSettings::showSuggestions() )
		keyboardLyt->addWidget( suggestions, 0, 0, 1, 2 );

    keyStack = new QStackedWidget();
    keyStack->setContentsMargins( QMargins() );

	Q_FOREACH( QString page, keymap->value( "Pages" ).toStringList() ) {
		QStringList rows = keymap->value( "Rows" ).toStringList();

		QVBoxLayout *pageLyt = new QVBoxLayout();
		pageLyt->setContentsMargins( QMargins() );
		pageLyt->setSpacing( 0 );

		pageLyt->addStretch();

		Q_FOREACH( QString row, rows ) {

			QHBoxLayout *rowLyt = new QHBoxLayout();
			rowLyt->setContentsMargins( QMargins() );
			rowLyt->setSpacing( 0 );

			Q_FOREACH( QString chr, keymap->value( page + "/" + row ).toStringList() ) {
				unsigned int keyCode = chr.toUInt();
				QString keyName1, keyName2;

				if ( keyCode == 0 ) {
					rowLyt->addStretch();
					continue;
				}

				if ( chr.startsWith( "000" ) ) {
					keyName1 = keymap->value( "Data/" + chr.replace( "000", "" ) ).toStringList().value( 0 );
					keyName2 = keymap->value( "Data/" + chr.replace( "000", "" ) ).toStringList().value( 1 );
				}

				else if ( chr.startsWith( "00" ) ) {
					keyName1 = keymap->value( "Data/" + chr.replace( "00", "" ) ).toStringList().value( 0 );
					keyName2 = keymap->value( "Data/" + chr.replace( "00", "" ) ).toStringList().value( 1 );
				}

				else if ( chr.startsWith( "50" ) ) {
					keyName1 = keymap->value( "Data/" + chr.replace( "50", "" ) ).toStringList().value( 1 );
					keyName2 = "";
				}

				else if ( chr.startsWith( "0" ) ) {
					keyName1 = keymap->value( "Data/" + chr.replace( 0, 1, "" ) ).toStringList().value( 0 );
					keyName2 = keymap->value( "Data/" + chr.replace( 0, 1, "" ) ).toStringList().value( 1 );
				}

				Button *btn = new Button( keyCode, keyName1, keyName2, this );
				rowLyt->addWidget( btn );

				/* Modifiers: Shift */
				if ( ( keyCode == 50 ) or ( keyCode == 62 ) ) {
					modifiers[ keyCode ] = btn;
					connect( btn, SIGNAL( longPressed() ), this, SLOT( shiftClicked() ) );
					connect( btn, SIGNAL( clicked() ), this, SLOT( shiftClicked() ) );
				}

				/* Modifiers: Ctrl */
				else if ( ( keyCode == 37 ) or ( keyCode == 109 ) ) {
					modifiers[ keyCode ] = btn;
					// connect( btn, SIGNAL( longPressed() ), this, SLOT( shiftClicked() ) );
					// connect( btn, SIGNAL( clicked() ), this, SLOT( shiftClicked() ) );
				}

				/* Modifiers: Alt */
				else if ( ( keyCode == 64 ) or ( keyCode == 108 ) ) {
					modifiers[ keyCode ] = btn;
					// connect( btn, SIGNAL( longPressed() ), this, SLOT( shiftClicked() ) );
					// connect( btn, SIGNAL( clicked() ), this, SLOT( shiftClicked() ) );
				}

				/* Modifiers: Meta */
				else if ( ( keyCode == 133 ) ) {
					modifiers[ keyCode ] = btn;
					// connect( btn, SIGNAL( longPressed() ), this, SLOT( shiftClicked() ) );
					// connect( btn, SIGNAL( clicked() ), this, SLOT( shiftClicked() ) );
				}

				if ( keyCode == 888 ) {
					connect( btn, SIGNAL( clicked() ), this, SLOT( switchPage() ) );
					connect( btn, SIGNAL( longPressed() ), this, SLOT( toggleShowHide() ) );
				}

				else {

					connect( btn, SIGNAL( sendKey( unsigned int ) ), this, SLOT( processKeyPress( unsigned int ) ) );
				}
			}

			pageLyt->addLayout( rowLyt );
		}

		pageLyt->addStretch();

		QWidget *pageW = new QWidget();
		pageW->setLayout( pageLyt );

		keyStack->addWidget( pageW );
	}

	keyboardLyt->addWidget( keyStack, 1, 0 );

	if ( CKSettings::keyboardLayout() == CKSettings::Desktop and CKSettings::showExtraDesktopKeys() ) {
		QStringList rows = keymap->value( "Rows" ).toStringList();

		QVBoxLayout *pageLyt = new QVBoxLayout();
		pageLyt->setContentsMargins( QMargins() );
		pageLyt->setSpacing( 0 );

		Q_FOREACH( QString row, rows ) {

			QHBoxLayout *rowLyt = new QHBoxLayout();
			rowLyt->setContentsMargins( QMargins() );
			rowLyt->setSpacing( 0 );

			Q_FOREACH( QString chr, keymap->value( "Extra/" + row ).toStringList() ) {
				unsigned int keyCode = chr.toUInt();
				QString keyName1, keyName2;

				if ( keyCode == 0 ) {
					rowLyt->addStretch();
					continue;
				}

				if ( chr.startsWith( "000" ) ) {
					keyName1 = keymap->value( "Data/" + chr.replace( "000", "" ) ).toStringList().value( 0 );
					keyName2 = keymap->value( "Data/" + chr.replace( "000", "" ) ).toStringList().value( 1 );
				}

				else if ( chr.startsWith( "00" ) ) {
					keyName1 = keymap->value( "Data/" + chr.replace( "00", "" ) ).toStringList().value( 0 );
					keyName2 = keymap->value( "Data/" + chr.replace( "00", "" ) ).toStringList().value( 1 );
				}

				else if ( chr.startsWith( "50" ) ) {
					keyName1 = keymap->value( "Data/" + chr.replace( "50", "" ) ).toStringList().value( 1 );
					keyName2 = "";
				}

				else if ( chr.startsWith( "0" ) ) {
					keyName1 = keymap->value( "Data/" + chr.replace( 0, 1, "" ) ).toStringList().value( 0 );
					keyName2 = keymap->value( "Data/" + chr.replace( 0, 1, "" ) ).toStringList().value( 1 );
				}

				else {
					keyName1 = keymap->value( "Data/" + chr ).toStringList().value( 0 );
					keyName2 = keymap->value( "Data/" + chr ).toStringList().value( 1 );
				}

				Button *btn = new Button( keyCode, keyName1, keyName2, this );
				rowLyt->addWidget( btn );

				if ( keyCode == 1 )
					btn->setDisabled( true );

				connect( btn, SIGNAL( sendKey( unsigned int ) ), this, SLOT( processKeyPress( unsigned int ) ) );
			}

			pageLyt->addLayout( rowLyt );
		}

		pageLyt->addStretch();

		QWidget *pageW = new QWidget();
		pageW->setLayout( pageLyt );

		keyboardLyt->addWidget( pageW, 1, 1 );
	}

	QWidget *base = new QWidget();
    base->setLayout( keyboardLyt );

    setCentralWidget( base );
};

void CoreKeyboard::setWidgetProperties() {

    setWindowTitle( "CoreKeyboard" );
    setWindowIcon( QIcon(":/resources/corekeyboard.svg") );

    setAttribute( Qt::WA_ShowWithoutActivating );
	setWindowFlags( Qt::FramelessWindowHint | Qt::WindowStaysOnTopHint | Qt::WindowDoesNotAcceptFocus );

	QSize kbdSize = QSize( MINIMUM_BUTTON_SIZE * maxCols + maxCols + 1, MINIMUM_BUTTON_SIZE * maxRows );

    QRect scrnSize = qApp->primaryScreen()->availableGeometry();
    int hpos = qApp->primaryScreen()->availableGeometry().x();
    int vpos = scrnSize.height() - kbdSize.height();

	setFixedHeight( MINIMUM_BUTTON_SIZE * maxRows );

	/* Full width */
	if ( CKSettings::keyboardMode() == CKSettings::Fixed ) {
		setFixedWidth( scrnSize.width() );
		setAttribute( Qt::WA_X11NetWmWindowTypeDock );
	}

	/* Compact */
	else
		setFixedWidth( MINIMUM_BUTTON_SIZE * maxCols * 1.5 + maxCols + 1 );

    move( hpos, vpos );
};

void CoreKeyboard::sendKey( unsigned int keyCode ) {

    Window currentFocus;
    int revertTo;

    Display *display = QX11Info::display();
    XGetInputFocus( display, &currentFocus, &revertTo );

    /* Toggle modes: Fixed <==> Compact */
	if ( keyCode == 22 ) {
		/* Get shift modifier */
		bool shiftPressed = false;
		Q_FOREACH( Button *btn, modifiers.values() ) {
			if ( ( btn->keyCode() == 50 ) or ( btn->keyCode() == 62 ) ) {
				shiftPressed |= ( btn->isChecked() and not btn->persist() );
				break;
			}
		}

		/* Swtch the mode if shift is pressed */
		if ( shiftPressed  )
			switchMode();
	}

	/* Only first page */
	if ( not keyStack->currentIndex() ) {
		/* Ctrl-[] actions, insert a Ctrl before actual key press*/
		if ( keyCode > 3700 ) {
			/* Insert Ctrl before sending the key */
			XTestFakeKeyEvent( display, 37, true, 2 );

			/* Virtual press of the actual key */
			XTestFakeKeyEvent( display, keyCode - 3700, true, 2 );
			XTestFakeKeyEvent( display, keyCode - 3700, false, 2 );

			/* Release shift after sending the key */
			XTestFakeKeyEvent( display, 37, false, 2 );
		}

		else {
			QListIterator<Button *> itr( modifiers.values() );
			while ( itr.hasNext() ) {
				Button *mod = itr.next();
				if ( mod->isChecked() ) {
					XTestFakeKeyEvent( display, mod->keyCode(), true, 2 );
				}
			}

			XTestFakeKeyEvent( display, keyCode, true, 2 );
			XTestFakeKeyEvent( display, keyCode, false, 2 );

			itr.toFront();
			while ( itr.hasNext() ) {
				Button *mod = itr.next();
				if ( mod->isChecked() )
					XTestFakeKeyEvent( display, mod->keyCode(), false, 2 );
			}
		}
	}

	/* Other pages */
	else {
		if ( keyCode > 5000 ) {
			/* Insert shift before sending the key */
			XTestFakeKeyEvent( display, 50, true, 2 );

			/* Virtual press of the actual key */
			XTestFakeKeyEvent( display, keyCode - 5000, true, 2 );
			XTestFakeKeyEvent( display, keyCode - 5000, false, 2 );

			/* Release shift after sending the key */
			XTestFakeKeyEvent( display, 50, false, 2 );
		}

		/* Ctrl-[] actions, insert a Ctrl before actual key press*/
		else if ( keyCode > 3700 ) {
			/* Insert shift before sending the key */
			XTestFakeKeyEvent( display, 37, true, 2 );

			/* Virtual press of the actual key */
			XTestFakeKeyEvent( display, keyCode - 3700, true, 2 );
			XTestFakeKeyEvent( display, keyCode - 3700, false, 2 );

			/* Release shift after sending the key */
			XTestFakeKeyEvent( display, 37, false, 2 );
		}

		else {
			XTestFakeKeyEvent( display, keyCode, true, 2 );
			XTestFakeKeyEvent( display, keyCode, false, 2 );
		}
	}

    XFlush( display );
};

void CoreKeyboard::switchMode() {

	close();

	sm->setValue( "CoreKeyboard", "Mode", CKSettings::keyboardMode() == CKSettings::Fixed ? 1 : 0 );
	setAttribute( Qt::WA_X11NetWmWindowTypeDock, CKSettings::keyboardMode() == CKSettings::Fixed );		/* Fixed = Not Movable = Dock */

	keymap = new QSettings( CKSettings::keyMapping(), QSettings::NativeFormat );

	maxRows = keymap->value( "Rows" ).toStringList().count() + 1;

	/* Dirty test hack */
	if ( keymap->value( "Pages" ).toStringList().count() == 1 )
		maxCols = 15;

	if ( CKSettings::keyboardLayout() == CKSettings::Desktop and CKSettings::showExtraDesktopKeys() )
		maxCols += 3;

    createKeyboard();
    setWidgetProperties();

    show();
};

void CoreKeyboard::updatePredictions() {

	QHBoxLayout *lyt = new QHBoxLayout();
	lyt->setContentsMargins( QMargins() );
	lyt->addStretch();

	if ( predictions.count() ) {
		/* At max, predict 20 words */
		for( int i = 0; i <= 20; i++ ) {
			if ( predictions.count() <= i )
				break;

			QString word( predictions[ i ] );

			QPushButton *btn = new QPushButton( word );
			btn->setFixedSize( fm->boundingRect( word ).width() + 10, 28 );
			btn->setFlat( true );
			btn->setFont( QFont( font().family(), 9, QFont::Normal ) );
			btn->setFocusPolicy( Qt::NoFocus );

			connect( btn, SIGNAL( clicked() ), this, SLOT( typeWord() ) );

			lyt->addWidget( btn );
		}
	}

	else {
		QLabel *lbl = new QLabel( "Start typing to see word predictions here..." );
		lyt->addWidget( lbl );
	}

	lyt->addStretch();

	QWidget *preds = new QWidget();
	preds->setLayout( lyt );

	suggestions->setWidget( preds );
};

void CoreKeyboard::processKeyPress( unsigned int keyCode ) {

    suggestWords( keyCode );
    sendKey( keyCode );

	if ( keymap->value( QString( "Data/%1" ).arg( keyCode ) ).toStringList().count() > 1 ) {
		/* Neutralize the modifier keys */
		QListIterator<Button *> itr( modifiers.values() );
		while ( itr.hasNext() ) {
			Button *mod = itr.next();
			if ( mod->isChecked() and not mod->persist() )
				mod->setChecked( false );
		}
	}
};

/* This is like pressing shift and then pressing @keyCode for all keys except bksp */
/* BKSP long press will be plain long press */
void CoreKeyboard::processLongKeyPress( unsigned int keyCode ) {

    suggestWords( keyCode );
    sendKey( keyCode );

	if ( keymap->value( QString( "Data/%1" ).arg( keyCode ) ).toStringList().count() > 1 ) {
		/* Neutralize the modifier keys */
		QListIterator<Button *> itr( modifiers.values() );
		while ( itr.hasNext() ) {
			Button *mod = itr.next();
			if ( mod->isChecked() and not mod->persist() )
				mod->setChecked( false );
		}
	}
};

void CoreKeyboard::toggleShowHide() {

    if ( isVisible() )
        hide();

    else
        show();
};

void CoreKeyboard::mousePressEvent( QMouseEvent *mpEvent ) {

    gpress = mpEvent->globalPos();

	if ( CKSettings::keyboardMode() == CKSettings::Compact ) {
		if ( QRect( 0, 0, width(), MINIMUM_BUTTON_SIZE ).contains( mpEvent->pos() ) )
			dragged = true;
	}

    dragPoint = mpEvent->pos();
};

void CoreKeyboard::mouseMoveEvent( QMouseEvent *mmEvent ) {

    if ( dragPoint.isNull() )
        return;

    QPoint curr( mmEvent->globalPos().x() - dragPoint.x(), mmEvent->globalPos().y() - dragPoint.y() );
    move( curr );
};

void CoreKeyboard::mouseReleaseEvent( QMouseEvent * ) {

    dragged = false;
    dragPoint = QPoint();
};

void CoreKeyboard::resizeEvent( QResizeEvent *rEvent ) {

	QMainWindow::resizeEvent( rEvent );

	int btnWidth = ( rEvent->size().width() - maxCols - 1 ) / maxCols;
	int btnHeight = ( rEvent->size().height() - MINIMUM_BUTTON_SIZE - maxRows - 2 ) / maxRows;

	Q_FOREACH( Button *btn, findChildren<Button *>() )
		btn->resize( btnWidth, btnHeight );
};

void CoreKeyboard::suggestWords( unsigned int keyCode ) {

	/* No auto-suggest when we have a modifier key other than shift, clicked */
	Q_FOREACH( Button *btn, modifiers.values() ) {
		if ( btn->isChecked() and ( btn->keyCode() != 50 and btn->keyCode() != 62 ) )
			return;
	}

	/* Back space */
	if ( keyCode == 22 ) {
		if ( typedLetters.count() >= 1 )
			typedLetters = typedLetters.left( typedLetters.count() - 1 );
	}

	/* Space, the word separator: Clears the typed letters */
	else if ( keyCode == 65 ) {
		typedLetters.clear();
		predictions.clear();
	}

	/* Add to the already typed letters */
	else {
		QString letter = keymap->value( QString( "Data/%1" ).arg( keyCode ) ).toString();
		if ( alphabet.contains( letter.toLower() ) )
			typedLetters += letter;

		else
			typedLetters.clear();
	}

	if ( typedLetters.count() )
		predictions = dict->wordsFor( typedLetters.toLower() );

	else
		predictions.clear();

	updatePredictions();
};

void CoreKeyboard::clearSuggestions() {

	typedLetters.clear();
	predictions.clear();

	updatePredictions();
};

void CoreKeyboard::shiftClicked() {

	Q_FOREACH( Button *btn, findChildren<Button *>() )
		btn->changeCase();
};

void CoreKeyboard::switchPage() {

	keyStack->setCurrentIndex( ( keyStack->currentIndex() + 1 ) % 2 );
};

void CoreKeyboard::typeWord() {

	QPushButton *btn = qobject_cast<QPushButton *>( sender() );
	if ( btn ) {
		QString text = btn->text().toLower();
		text = text.right( text.size() - typedLetters.size() );

		Display *display = QX11Info::display();

		for( int i = 0; i < text.count(); i++ ) {
			processKeyPress( XKeysymToKeycode( display, text.at( i ).cell() ) );
		}

		/* Insert a space */
		processKeyPress( XKeysymToKeycode( display, ' ' ) );
	}

	typedLetters.clear();

	clearSuggestions();
};
